
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* bxMath.cpp */


#include <math.h>
#include "bxMath.h"

using namespace NameBeziex;



BxVec2::BxVec2( const BxVec2& a )
{
	X() = a.X();
	Y() = a.Y();
}

BxVec2::BxVec2( const double x, const double y )
{
	X() = x;
	Y() = y;
}

BxVec2&  BxVec2::
operator+=( const BxVec2& a )
{
	X() += a.X();
	Y() += a.Y();

	return *this;
}

BxVec2&  BxVec2::
operator-=( const BxVec2& a )
{
	X() -= a.X();
	Y() -= a.Y();

	return *this;
}

BxVec2&  BxVec2::
operator*=( const double a )
{
	X() *= a;
	Y() *= a;

	return *this;
}

BxVec2&  BxVec2::
operator/=( const double a )
{
	X() /= a;
	Y() /= a;

	return *this;
}

const BxVec2  BxVec2::
operator+( const BxVec2& a ) const
{
	return BxVec2( X() + a.X(), Y() + a.Y() );
}

const BxVec2  BxVec2::
operator-( const BxVec2& a ) const
{
	return BxVec2( X() - a.X(), Y() - a.Y() );
}

const BxVec2  BxVec2::
operator/( const double a ) const
{
	return BxVec2( X() / a, Y() / a );
}

BxVec2
operator*( const BxVec2& a, const double b )
{
	return BxVec2( a.X() * b, a.Y() * b );
}

BxVec2
operator*( const double a, const BxVec2& b )
{
	return BxVec2( a * b.X(), a * b.Y() );
}

double
Dot( const BxVec2& a, const BxVec2& b )
{
	return a.X() * b.X() + a.Y() * b.Y();
}

double
Cross( const BxVec2& a, const BxVec2& b )
{
	return a.X() * b.Y() - a.Y() * b.X();
}

BxVec2
Arc( const BxVec2& vecPntToPnt, const BxVec2& unitTan, BxErr* errResult )
{
	BxErr  errResult0;

	if( errResult != 0 )
		*errResult = kBxErrSuccess;

	BxVec2  unitVecPntToPnt = vecPntToPnt.Normalize( &errResult0 );
	if( errResult0 == kBxErrFailure ) {
		if( errResult != 0 ) {
			*errResult = kBxErrFailure;
			return unitTan;
		}
	}

	double  cosDt  = Dot( unitVecPntToPnt, unitTan );
	double  sinDt2 = 1.0 - cosDt * cosDt;
	double  sinDt  = 0.0;
	if( sinDt2 > 0.0 )
		sinDt = sqrt( sinDt2 );

	if( -GBxNearZero() < sinDt && sinDt < GBxNearZero() )
		return vecPntToPnt / 3.0;

	return ( 2.0 * vecPntToPnt * ( 1.0 - cosDt ) ) / ( 3.0 * sinDt );
}

const double  BxVec2::
Length() const
{
	return sqrt( X() * X() + Y() * Y() );
}

const double  BxVec2::
LengthSquared() const
{
	return X() * X() + Y() * Y();
}

const BxVec2  BxVec2::
Normalize() const
{
	return Normalize( GBxZero2() );
}

const BxVec2  BxVec2::
Normalize( const BxVec2& resultForErr ) const
{
	double  len = Length();

	if( len < GBxNearZero() )
		return BxVec2( resultForErr );
	else
		return BxVec2( *this / len );
}

const BxVec2  BxVec2::
Normalize( BxErr* errResult ) const
{
	double  len = Length();

	if( len < GBxNearZero() ) {
		*errResult = kBxErrFailure;

		return BxVec2( GBxZero2() );
	}
	else {
		*errResult = kBxErrSuccess;

		return BxVec2( *this / len );
	}
}

const BxVec2  BxVec2::
operator-() const
{
	return BxVec2( -X(), -Y() );
}

const BxVec2  BxVec2::
Abs() const
{
	BxVec2  a;

	if( X() >= 0.0 )
		a.X() =  X();
	else
		a.X() = -X();

	if( Y() >= 0.0 )
		a.Y() =  Y();
	else
		a.Y() = -Y();

	return a;
}

const BxVec2  BxVec2::
Point( const BxMat2x3& a ) const
{
	BxVec2  b;

	b.X() = X() * a[ 0 ][ 0 ] + Y() * a[ 0 ][ 1 ] + a[ 0 ][ 2 ];
	b.Y() = X() * a[ 1 ][ 0 ] + Y() * a[ 1 ][ 1 ] + a[ 1 ][ 2 ];

	return b;
}

const BxVec2  BxVec2::
Vector( const BxMat2x3& a ) const
{
	BxVec2  b;

	b.X() = X() * a[ 0 ][ 0 ] + Y() * a[ 0 ][ 1 ];
	b.Y() = X() * a[ 1 ][ 0 ] + Y() * a[ 1 ][ 1 ];

	return b;
}



BxDir::BxDir( const BxDir& a )
	: BxVec2( a )
{
}

BxDir::BxDir( const BxVec2& a )
{
	X() = a.X();
	Y() = a.Y();
}

BxDir::BxDir( const double x, const double y )
	: BxVec2( x, y )
{
}

BxDir::BxDir( const double angle )
{
	X() = cos( angle );
	Y() = sin( angle );
}

const BxDir  BxDir::
operator-() const
{
	return BxDir( X(), -Y() );
}

const double  BxDir::
Sin( BxErr* errResult ) const
{
	BxErr  errResult0;

	BxDir  a = Normalize( &errResult0 );

	if( errResult != 0 )
		*errResult = errResult0;

	if( errResult0 == kBxErrFailure )
		return 0;

	return a.Y();
}

const double  BxDir::
Cos( BxErr* errResult ) const
{
	BxErr  errResult0;

	BxDir  a = Normalize( &errResult0 );

	if( errResult != 0 )
		*errResult = errResult0;

	if( errResult0 == kBxErrFailure )
		return 0;

	return a.X();
}

const double  BxDir::
SignedAngle( BxErr* errResult ) const
{
	BxErr  errResult0;

	BxDir  a = Normalize( &errResult0 );

	if( errResult != 0 )
		*errResult = errResult0;

	if( errResult0 == kBxErrFailure )
		return 0;

	if( fabs( a.Y() ) <= fabs( a.X() ) ) {
		if( a.X() >= 0.0 )
			return asin( a.Y() );
		else {
			if( a.Y() >= 0.0 )
				return  GBxPi() - asin( a.Y() );
			else
				return -GBxPi() - asin( a.Y() );
		}
	}
	else {
		if( a.Y() >= 0.0 )
			return  acos( a.X() );
		else
			return -acos( a.X() );
	}
}

const double  BxDir::
UnsignedAngle( BxErr* errResult ) const
{
	BxErr  errResult0;

	double  sign = SignedAngle( &errResult0 );

	if( errResult != 0 )
		*errResult = errResult0;

	if( errResult0 == kBxErrFailure )
		return sign;

	if( sign >= 0.0 )
		return sign;
	else
		return ( 2.0 * GBxPi() ) - sign;
}

bool  BxDir::
GreaterSignedAngle( const BxDir& cmpDt, BxErr* errResult )
{
	double  ang0, ang1;

	BxErr  errResult0 = SignedAngleSub( cmpDt, ang0, ang1 );

	if( errResult != 0 )
		*errResult = errResult0;

	if( errResult0 == kBxErrFailure )
		return false;

	if( ang0 > ang1 )
		return true;

	return false;
}

bool  BxDir::
LessSignedAngle( const BxDir& cmpDt, BxErr* errResult )
{
	double  ang0, ang1;

	BxErr  errResult0 = SignedAngleSub( cmpDt, ang0, ang1 );

	if( errResult != 0 )
		*errResult = errResult0;

	if( errResult0 == kBxErrFailure )
		return false;

	if( ang0 < ang1 )
		return true;

	return false;
}

BxErr  BxDir::
SignedAngleSub( const BxDir& cmpDt, double& ang0, double& ang1 )
{
	BxErr  errResult;

	ang0 = SignedAngle( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	ang1 = cmpDt.SignedAngle( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}

bool  BxDir::
GreaterUnsignedAngle( const BxDir& cmpDt, BxErr* errResult )
{
	double  ang0, ang1;

	BxErr  errResult0 = UnsignedAngleSub( cmpDt, ang0, ang1 );

	if( errResult != 0 )
		*errResult = errResult0;

	if( errResult0 == kBxErrFailure )
		return false;

	if( ang0 > ang1 )
		return true;

	return false;
}

bool  BxDir::
LessUnsignedAngle( const BxDir& cmpDt, BxErr* errResult )
{
	double  ang0, ang1;

	BxErr  errResult0 = UnsignedAngleSub( cmpDt, ang0, ang1 );

	if( errResult != 0 )
		*errResult = errResult0;

	if( errResult0 == kBxErrFailure )
		return false;

	if( ang0 < ang1 )
		return true;

	return false;
}

BxErr  BxDir::
UnsignedAngleSub( const BxDir& cmpDt, double& ang0, double& ang1 )
{
	BxErr  errResult;

	ang0 = UnsignedAngle( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	ang1 = cmpDt.UnsignedAngle( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}




BxVec3::BxVec3( const BxVec3& a )
{
	X() = a.X();
	Y() = a.Y();
	Z() = a.Z();
}

BxVec3::BxVec3( const double x, const double y, const double z )
{
	X() = x;
	Y() = y;
	Z() = z;
}

BxVec3&  BxVec3::
operator+=( const BxVec3& a )
{
	X() += a.X();
	Y() += a.Y();
	Z() += a.Z();

	return *this;
}

BxVec3&  BxVec3::
operator-=( const BxVec3& a )
{
	X() -= a.X();
	Y() -= a.Y();
	Z() -= a.Z();

	return *this;
}

BxVec3&  BxVec3::
operator*=( const double a )
{
	X() *= a;
	Y() *= a;
	Z() *= a;

	return *this;
}

BxVec3&  BxVec3::
operator/=( const double a )
{
	X() /= a;
	Y() /= a;
	Z() /= a;

	return *this;
}

const BxVec3  BxVec3::
operator+( const BxVec3& a ) const
{
	return BxVec3( X() + a.X(), Y() + a.Y(), Z() + a.Z() );
}

const BxVec3  BxVec3::
operator-( const BxVec3& a ) const
{
	return BxVec3( X() - a.X(), Y() - a.Y(), Z() - a.Z() );
}

const BxVec3  BxVec3::
operator/( const double a ) const
{
	return BxVec3( X() / a, Y() / a, Z() / a );
}

BxVec3
operator*( const BxVec3& a, const double b )
{
	return BxVec3( a.X() * b, a.Y() * b, a.Z() * b );
}

BxVec3
operator*( const double a, const BxVec3& b )
{
	return BxVec3( a * b.X(), a * b.Y(), a * b.Z() );
}

double
Dot( const BxVec3& a, const BxVec3& b )
{
	return a.X() * b.X() + a.Y() * b.Y() + a.Z() * b.Z();
}

BxVec3
Cross( const BxVec3& a, const BxVec3& b )
{
	return BxVec3(
		a.Y() * b.Z() - a.Z() * b.Y(),
		a.Z() * b.X() - a.X() * b.Z(),
		a.X() * b.Y() - a.Y() * b.X() );
}

BxVec3
Arc( const BxVec3& vecPntToPnt, const BxVec3& unitTan, BxErr* errResult )
{
	if( errResult != 0 )
		*errResult = kBxErrSuccess;

	double  lenVecPntToPnt = vecPntToPnt.Length();
	if( lenVecPntToPnt < GBxNearZero() ) {
		if( errResult != 0 ) {
			*errResult = kBxErrFailure;
			return unitTan;
		}
	}
	BxVec3  unitVecPntToPnt = vecPntToPnt / lenVecPntToPnt;

	double  cosDt  = Dot( unitVecPntToPnt, unitTan );
	double  sinDt2 = 1.0 - cosDt * cosDt;
	double  sinDt  = 0.0;
	if( sinDt2 > 0.0 )
		sinDt = sqrt( sinDt2 );

	if( -GBxNearZero() < sinDt && sinDt < GBxNearZero() )
		return vecPntToPnt / 3.0;

	double  lenTan = ( 2.0 * lenVecPntToPnt * ( 1.0 - cosDt ) ) / ( 3.0 * sinDt * sinDt );

	return lenTan * unitTan;
}

const double  BxVec3::
Length() const
{
	return sqrt( X() * X() + Y() * Y() + Z() * Z() );
}

const double  BxVec3::
LengthSquared() const
{
	return X() * X() + Y() * Y() + Z() * Z();
}

const BxVec3  BxVec3::
Normalize() const
{
	return Normalize( GBxZero3() );
}

const BxVec3  BxVec3::
Normalize( const BxVec3& resultForErr ) const
{
	double  len = Length();

	if( len < GBxNearZero() )
		return BxVec3( resultForErr );
	else
		return BxVec3( *this / len );
}

const BxVec3  BxVec3::
Normalize( BxErr* errResult ) const
{
	double  len = Length();

	if( len < GBxNearZero() ) {
		*errResult = kBxErrFailure;

		return BxVec3( GBxZero3() );
	}
	else {
		*errResult = kBxErrSuccess;

		return BxVec3( *this / len );
	}
}

const BxVec3  BxVec3::
operator-() const
{
	return BxVec3( -X(), -Y(), -Z() );
}

const BxVec3  BxVec3::
Abs() const
{
	BxVec3  a;

	if( X() >= 0.0 )
		a.X() =  X();
	else
		a.X() = -X();

	if( Y() >= 0.0 )
		a.Y() =  Y();
	else
		a.Y() = -Y();

	if( Z() >= 0.0 )
		a.Z() =  Z();
	else
		a.Z() = -Z();

	return a;
}

const BxVec3  BxVec3::
Point( const BxMat3x4& a ) const
{
	BxVec3  b;

	b.X() = X() * a[ 0 ][ 0 ] + Y() * a[ 0 ][ 1 ] + Z() * a[ 0 ][ 2 ] + a[ 0 ][ 3 ];
	b.Y() = X() * a[ 1 ][ 0 ] + Y() * a[ 1 ][ 1 ] + Z() * a[ 1 ][ 2 ] + a[ 1 ][ 3 ];
	b.Z() = X() * a[ 2 ][ 0 ] + Y() * a[ 2 ][ 1 ] + Z() * a[ 2 ][ 2 ] + a[ 2 ][ 3 ];

	return b;
}

const BxVec3  BxVec3::
Vector( const BxMat3x4& a ) const
{
	BxVec3  b;

	b.X() = X() * a[ 0 ][ 0 ] + Y() * a[ 0 ][ 1 ] + Z() * a[ 0 ][ 2 ];
	b.Y() = X() * a[ 1 ][ 0 ] + Y() * a[ 1 ][ 1 ] + Z() * a[ 1 ][ 2 ];
	b.Z() = X() * a[ 2 ][ 0 ] + Y() * a[ 2 ][ 1 ] + Z() * a[ 2 ][ 2 ];

	return b;
}



BxAxis::BxAxis( const BxAxis& a )
{
	MainAxis() = a.MainAxis();
	SubAxis()  = a.SubAxis();
}

BxAxis::BxAxis( const BxVec3& mainAxis, const BxVec3& subAxis )
{
	MainAxis() = mainAxis;
	SubAxis()  = subAxis;
}

const BxAxis  BxAxis::
RegularAxis( BxErr* errResult ) const
{
	BxAxis   z;
	BxVec3   a;
	BxErr    errResult0;

	if( errResult != 0 )
		*errResult = kBxErrSuccess;

	z.MainAxis() = MainAxis().Normalize( &errResult0 );
	if( errResult0 == kBxErrFailure ) {
		if( errResult != 0 )
			*errResult = kBxErrFailure;
		return z;
	}

	a = Cross( Cross( z.MainAxis(), SubAxis() ), z.MainAxis() );
	z.SubAxis() = a.Normalize( &errResult0 );
	if( errResult0 == kBxErrFailure ) {
		if( errResult != 0 )
			*errResult = kBxErrFailure;
		return z;
	}

	return z;
}

BxAxis
MakeSubAxis( const BxVec3& mainAxis, BxErr* errResult )
{
	short  kindAxis;
	BxErr  errResult0;

	if( errResult != 0 )
		*errResult = kBxErrSuccess;

	BxVec3  a = mainAxis.Abs();
	if( a.X() >= a.Y() ) {
		if( a.X() >= a.Z() )
			kindAxis = BxVec3::kElmX;
		else
			kindAxis = BxVec3::kElmZ;
	}
	else {
		if( a.Y() >= a.Z() )
			kindAxis = BxVec3::kElmY;
		else
			kindAxis = BxVec3::kElmZ;
	}

	switch( kindAxis ) {
		case  BxVec3::kElmX:
			a = GBxWorldAxis().AxisY();
			break;
		case  BxVec3::kElmY:
			a = GBxWorldAxis().AxisZ();
			break;
		case  BxVec3::kElmZ:
			a = GBxWorldAxis().AxisX();
			break;
	}

	BxAxis  b( mainAxis, a );
	BxAxis  result = b.RegularAxis( &errResult0 );
	if( errResult0 == kBxErrFailure ) {
		if( errResult != 0 )
			*errResult = kBxErrFailure;
		return result;
	}

	return result;
}

const BxVec3  BxAxis::
GetAxisZ( BxErr* errResult ) const
{
	BxErr  errResult0;

	if( errResult != 0 )
		*errResult = kBxErrSuccess;

	BxAxis  a = RegularAxis( &errResult0 );
	if( errResult0 == kBxErrFailure ) {
		if( errResult != 0 )
			*errResult = kBxErrFailure;
		return BxVec3();
	}

	return Cross( a.MainAxis(), a.SubAxis() );
}

const BxAxisXYZ  BxAxis::
Conv( BxErr* errResult ) const
{
	BxAxisXYZ  result;
	BxErr      errResult0;

	if( errResult != 0 )
		*errResult = kBxErrSuccess;

	BxAxis  a = RegularAxis( &errResult0 );
	if( errResult0 == kBxErrFailure ) {
		if( errResult != 0 )
			*errResult = kBxErrFailure;
		return result;
	}

	result.AxisX() = a.MainAxis();
	result.AxisY() = a.SubAxis();
	result.AxisZ() = Cross( result.AxisX(), result.AxisY() );

	return result;
}



BxAxisXYZ::BxAxisXYZ( const BxAxisXYZ& a )
{
	AxisX() = a.AxisX();
	AxisY() = a.AxisY();
	AxisZ() = a.AxisZ();
}

BxAxisXYZ::BxAxisXYZ( const BxVec3& axisX, const BxVec3& axisY, const BxVec3& axisZ )
{
	AxisX() = axisX;
	AxisY() = axisY;
	AxisZ() = axisZ;
}



BxMat2x3::BxMat2x3()
{
	fDt[0][0] = 1.0;	fDt[1][0] = 0.0;
	fDt[0][1] = 0.0;	fDt[1][1] = 1.0;
	fDt[0][2] = 0.0;	fDt[1][2] = 0.0;
}

BxMat2x3::BxMat2x3( const BxMat2x3& a )
{
	for( int u=0; u<2; u++ ) {
		for( int v=0; v<3; v++ )
			fDt[ u ][ v ] = a.fDt[ u ][ v ];
	}
}

BxMat2x3::BxMat2x3( const double a[ 2 ][ 3 ] )
{
	for( int u=0; u<2; u++ ) {
		for( int v=0; v<3; v++ )
			fDt[ u ][ v ] = a[ u ][ v ];
	}
}

BxMat2x3::BxMat2x3( const double a00, const double a10,
	            const double a01, const double a11,
		    const double a02, const double a12 )
{
	fDt[0][0] = a00;	fDt[1][0] = a10;
	fDt[0][1] = a01;	fDt[1][1] = a11;
	fDt[0][2] = a02;	fDt[1][2] = a12;
}

BxMat2x3
operator*( const BxMat2x3& a, const BxMat2x3& b )
{
	int       u;
	BxMat2x3  c;

	for( int v=0; v<2; v++ ) {
		for( u=0; u<2; u++ ) {
			c.fDt[ u ][ v ] = a.fDt[ 0 ][ v ] * b.fDt[ u ][ 0 ]
			                + a.fDt[ 1 ][ v ] * b.fDt[ u ][ 1 ];
		}
	}

	for( u=0; u<2; u++ ) {
		c.fDt[ u ][ 2 ] = a.fDt[ 0 ][ 2 ] * b.fDt[ u ][ 0 ]
		                + a.fDt[ 1 ][ 2 ] * b.fDt[ u ][ 1 ]
		                +                   b.fDt[ u ][ 2 ];
	}

	return c;
}

BxMat2x3
Translete( const double x, const double y )
{
	BxMat2x3  a;

	a.fDt[ 0 ][ 2 ] = x;
	a.fDt[ 1 ][ 2 ] = y;

	return a;
}

BxMat2x3
Scale( const double x, const double y )
{
	BxMat2x3  a;

	a.fDt[ 0 ][ 0 ] = x;
	a.fDt[ 1 ][ 1 ] = y;

	return a;
}

BxMat2x3
Rotate( const double a )
{
	BxMat2x3  b;

	double  dtSin = sin( a );
	double  dtCos = cos( a );

	return b.RotateSub( dtSin, dtCos );
}

BxMat2x3
Rotate( const BxDir& a, BxErr* errResult )
{
	BxMat2x3  b;
	BxErr     errResult0, errResult1;

	double  dtSin = a.Sin( &errResult0 );
	double  dtCos = a.Cos( &errResult1 );

	if( errResult != 0 )
		*errResult = kBxErrSuccess;

	if( errResult0 == kBxErrFailure || errResult1 == kBxErrFailure ) {
		if( errResult != 0 )
			*errResult = kBxErrFailure;
		return b;
	}

	return b.RotateSub( dtSin, dtCos );
}

BxMat2x3&  BxMat2x3::
RotateSub( const double dtSin, const double dtCos )
{
	fDt[ 0 ][ 0 ] =  dtCos;		fDt[ 1 ][ 0 ] =  dtSin;
	fDt[ 0 ][ 1 ] = -dtSin;		fDt[ 1 ][ 1 ] =  dtCos;

	return *this;
}




BxMat3x4::BxMat3x4()
{
	fDt[0][0] = 1.0;	fDt[1][0] = 0.0;	fDt[2][0] = 0.0;
	fDt[0][1] = 0.0;	fDt[1][1] = 1.0;	fDt[2][1] = 0.0;
	fDt[0][2] = 0.0;	fDt[1][2] = 0.0;	fDt[2][2] = 1.0;
	fDt[0][3] = 0.0;	fDt[1][3] = 0.0;	fDt[2][3] = 0.0;
}

BxMat3x4::BxMat3x4( const BxMat3x4& a )
{
	for( int u=0; u<3; u++ ) {
		for( int v=0; v<4; v++ )
			fDt[ u ][ v ] = a.fDt[ u ][ v ];
	}
}

BxMat3x4::BxMat3x4( const double a[ 3 ][ 4 ] )
{
	for( int u=0; u<3; u++ ) {
		for( int v=0; v<4; v++ )
			fDt[ u ][ v ] = a[ u ][ v ];
	}
}

BxMat3x4::BxMat3x4( const double a00, const double a10, const double a20,
	            const double a01, const double a11, const double a21,
		    const double a02, const double a12, const double a22,
		    const double a03, const double a13, const double a23 )
{
	fDt[0][0] = a00;	fDt[1][0] = a10;	fDt[2][0] = a20;
	fDt[0][1] = a01;	fDt[1][1] = a11;	fDt[2][1] = a21;
	fDt[0][2] = a02;	fDt[1][2] = a12;	fDt[2][2] = a22;
	fDt[0][3] = a03;	fDt[1][3] = a13;	fDt[2][3] = a23;
}

BxMat3x4
operator*( const BxMat3x4& a, const BxMat3x4& b )
{
	int       u;
	BxMat3x4  c;

	for( int v=0; v<3; v++ ) {
		for( u=0; u<3; u++ ) {
			c.fDt[ u ][ v ] = a.fDt[ 0 ][ v ] * b.fDt[ u ][ 0 ]
			                + a.fDt[ 1 ][ v ] * b.fDt[ u ][ 1 ]
			                + a.fDt[ 2 ][ v ] * b.fDt[ u ][ 2 ];
		}
	}

	for( u=0; u<3; u++ ) {
		c.fDt[ u ][ 3 ] = a.fDt[ 0 ][ 3 ] * b.fDt[ u ][ 0 ]
		                + a.fDt[ 1 ][ 3 ] * b.fDt[ u ][ 1 ]
		                + a.fDt[ 2 ][ 3 ] * b.fDt[ u ][ 2 ]
		                +                   b.fDt[ u ][ 3 ];
	}

	return c;
}

BxMat3x4
Translete( const double x, const double y, const double z )
{
	BxMat3x4  a;

	a.fDt[ 0 ][ 3 ] = x;
	a.fDt[ 1 ][ 3 ] = y;
	a.fDt[ 2 ][ 3 ] = z;

	return a;
}

BxMat3x4
Scale( const double x, const double y, const double z )
{
	BxMat3x4  a;

	a.fDt[ 0 ][ 0 ] = x;
	a.fDt[ 1 ][ 1 ] = y;
	a.fDt[ 2 ][ 2 ] = z;

	return a;
}

BxMat3x4
RotateX( const double x )
{
	BxMat3x4  a;

	double  dtSin = sin( x );
	double  dtCos = cos( x );

	return a.RotateXSub( dtSin, dtCos );
}

BxMat3x4
RotateY( const double y )
{
	BxMat3x4  a;

	double  dtSin = sin( y );
	double  dtCos = cos( y );

	return a.RotateYSub( dtSin, dtCos );
}

BxMat3x4
RotateZ( const double z )
{
	BxMat3x4  a;

	double  dtSin = sin( z );
	double  dtCos = cos( z );

	return a.RotateZSub( dtSin, dtCos );
}

BxMat3x4
RotateAxis( const double angle, const BxVec3& axis, BxErr* errResult )
{
	BxMat3x4  a;
	BxErr     errResult0;

	double  dtSin = sin( angle );
	double  dtCos = cos( angle );

	a.RotateAxisSub( dtSin, dtCos, axis, errResult0 );

	if( errResult != 0 )
		*errResult = errResult0;

	return a;
}

BxMat3x4
Scale( const BxDir& a, BxErr* errResult )
{
	if( errResult != 0 )
		*errResult = kBxErrSuccess;

	double  scale = a.Length();
	if( scale < GBxNearZero() ) {
		if( errResult != 0 )
			*errResult = kBxErrFailure;
		return BxMat3x4();
	}

	return Scale( scale, scale, scale );
}

BxMat3x4
RotateX( const BxDir& a, BxErr* errResult )
{
	BxMat3x4  b;
	BxErr     errResult0, errResult1;

	double  dtSin = a.Sin( &errResult0 );
	double  dtCos = a.Cos( &errResult1 );

	if( errResult != 0 )
		*errResult = kBxErrSuccess;

	if( errResult0 == kBxErrFailure || errResult1 == kBxErrFailure ) {
		if( errResult != 0 )
			*errResult = kBxErrFailure;
		return b;
	}

	return b.RotateXSub( dtSin, dtCos );
}

BxMat3x4
RotateY( const BxDir& a, BxErr* errResult )
{
	BxMat3x4  b;
	BxErr     errResult0, errResult1;

	double  dtSin = a.Sin( &errResult0 );
	double  dtCos = a.Cos( &errResult1 );

	if( errResult != 0 )
		*errResult = kBxErrSuccess;

	if( errResult0 == kBxErrFailure || errResult1 == kBxErrFailure ) {
		if( errResult != 0 )
			*errResult = kBxErrFailure;
		return b;
	}

	return b.RotateYSub( dtSin, dtCos );
}

BxMat3x4
RotateZ( const BxDir& a, BxErr* errResult )
{
	BxMat3x4  b;
	BxErr     errResult0, errResult1;

	double  dtSin = a.Sin( &errResult0 );
	double  dtCos = a.Cos( &errResult1 );

	if( errResult != 0 )
		*errResult = kBxErrSuccess;

	if( errResult0 == kBxErrFailure || errResult1 == kBxErrFailure ) {
		if( errResult != 0 )
			*errResult = kBxErrFailure;
		return b;
	}

	return b.RotateZSub( dtSin, dtCos );
}

BxMat3x4
RotateAxis( const BxDir& a, const BxVec3& axis, BxErr* errResult )
{
	BxMat3x4  b;
	BxErr     errResult0, errResult1;

	double  dtSin = a.Sin( &errResult0 );
	double  dtCos = a.Cos( &errResult1 );

	if( errResult != 0 )
		*errResult = kBxErrSuccess;

	if( errResult0 == kBxErrFailure || errResult1 == kBxErrFailure ) {
		if( errResult != 0 )
			*errResult = kBxErrFailure;
		return b;
	}

	b.RotateAxisSub( dtSin, dtCos, axis, errResult0 );

	if( errResult != 0 )
		*errResult = errResult0;

	return b;
}

BxMat3x4&  BxMat3x4::
RotateXSub( const double dtSin, const double dtCos )
{
	fDt[ 1 ][ 1 ] =  dtCos;		fDt[ 2 ][ 1 ] =  dtSin;
	fDt[ 1 ][ 2 ] = -dtSin;		fDt[ 2 ][ 2 ] =  dtCos;

	return *this;
}

BxMat3x4&  BxMat3x4::
RotateYSub( const double dtSin, const double dtCos )
{
	fDt[ 2 ][ 2 ] =  dtCos;		fDt[ 0 ][ 2 ] =  dtSin;
	fDt[ 2 ][ 0 ] = -dtSin;		fDt[ 0 ][ 0 ] =  dtCos;

	return *this;
}

BxMat3x4&  BxMat3x4::
RotateZSub( const double dtSin, const double dtCos )
{
	fDt[ 0 ][ 0 ] =  dtCos;		fDt[ 1 ][ 0 ] =  dtSin;
	fDt[ 0 ][ 1 ] = -dtSin;		fDt[ 1 ][ 1 ] =  dtCos;

	return *this;
}

BxMat3x4&  BxMat3x4::
RotateAxisSub( const double dtSin, const double dtCos, const BxVec3& axis, BxErr& errResult )
{
	errResult = kBxErrSuccess;
	
	double  len = axis.Length();
	if( len < GBxNearZero() ) {
		errResult = kBxErrFailure;
		return *this;
	}
	BxVec3  axis2 = axis / len;

	double  pSin = dtSin;
	double  pCos = dtCos;
	double  mCos = 1.0 - pCos;

	fDt[ 0 ][ 0 ] =  pCos;
	fDt[ 1 ][ 0 ] =  pSin * axis2.Z();
	fDt[ 2 ][ 0 ] = -pSin * axis2.Y();

	fDt[ 0 ][ 1 ] = -pSin * axis2.Z();
	fDt[ 1 ][ 1 ] =  pCos;
	fDt[ 2 ][ 1 ] =  pSin * axis2.X();

	fDt[ 0 ][ 2 ] =  pSin * axis2.Y();
	fDt[ 1 ][ 2 ] = -pSin * axis2.X();
	fDt[ 2 ][ 2 ] =  pCos;

	for( int u=0; u<3; u++ ) {
		for( int v=0; v<3; v++ )
			fDt[ u ][ v ] += mCos * axis2[ u ] * axis2[ v ];
	}

	return *this;
}

BxMat3x4
RotWorldToLocal( const BxAxis& axis, BxErr* errResult )
{
	BxErr  errResult0;

	if( errResult != 0 )
		*errResult = kBxErrSuccess;

	BxAxisXYZ  dstAxis = axis.Conv( &errResult0 );
	if( errResult0 == kBxErrFailure ) {
		if( errResult != 0 )
			*errResult = kBxErrFailure;
		return BxMat3x4();
	}

	return RotWorldLocalSub( GBxWorldAxis(), dstAxis );
}

BxMat3x4
RotLocalToWorld( const BxAxis& axis, BxErr* errResult )
{
	BxErr  errResult0;

	if( errResult != 0 )
		*errResult = kBxErrSuccess;

	BxAxisXYZ  srcAxis = axis.Conv( &errResult0 );
	if( errResult0 == kBxErrFailure ) {
		if( errResult != 0 )
			*errResult = kBxErrFailure;
		return BxMat3x4();
	}

	return RotWorldLocalSub( srcAxis, GBxWorldAxis() );
}

BxMat3x4
RotWorldLocalSub( const BxAxisXYZ& srcAxis, const BxAxisXYZ& dstAxis )
{
	BxMat3x4  result;
	int       u;

	for( int v=0; v<3; v++ ) {
		for( u=0; u<3; u++ )
			result.fDt[ u ][ v ] = Dot( srcAxis[ u ], dstAxis[ v ] );
	}

	for( u=0; u<3; u++ )
		result.fDt[ u ][ 3 ] = 0.0;

	return result;
}

BxMat3x4
RotateAxis( const BxAxis& srcAxis, const BxAxis& dstAxis, BxErr* errResult )
{
	BxErr  errResult0;

	if( errResult != 0 )
		*errResult = kBxErrSuccess;

	BxMat3x4  mat0 = RotLocalToWorld( srcAxis, &errResult0 );
	if( errResult0 == kBxErrFailure ) {
		if( errResult != 0 )
			*errResult = kBxErrFailure;
		return BxMat3x4();
	}

	BxMat3x4  mat1 = RotWorldToLocal( dstAxis, &errResult0 );
	if( errResult0 == kBxErrFailure ) {
		if( errResult != 0 )
			*errResult = kBxErrFailure;
		return BxMat3x4();
	}
	
	return ( mat0 * mat1 );
}



Bezier3Line::Bezier3Line( const Bezier3Line& a )
{
	for( int i=0; i<4; i++ )
		fDt[ i ] = a[ i ];
}

Bezier3Line::Bezier3Line( const BxVecAry4& a )
{
	for( int i=0; i<4; i++ )
		fDt[ i ] = a.p[ i ];
}

Bezier3Line::Bezier3Line( const BxVec3 a[ 4 ] )
{
	for( int i=0; i<4; i++ )
		fDt[ i ] = a[ i ];
}

BxVec3
Pos( const Bezier3Line& bezierLine, const double t )
{
	BxVec3  pos;

	double  t2  = t  * t;
	double  t3  = t2 * t;
	double  mt  = 1.0 - t;
	double  mt2 = mt  * mt;
	double  mt3 = mt2 * mt;

	for( int i=0; i<BxVec3::kNumVec3; i++ ) {
		pos[ i ] =            mt3 * bezierLine[ 0 ][ i ]
			 + 3.0 * t  * mt2 * bezierLine[ 1 ][ i ]
			 + 3.0 * t2 * mt  * bezierLine[ 2 ][ i ]
			 +       t3       * bezierLine[ 3 ][ i ];
	}

	return pos;
}

BxVec3
Diff( const Bezier3Line& bezierLine, const double t )
{
	BxVec3  diff;

	double  t2  = t * t;
	double  mt  = 1.0 - t;
	double  mt2 = mt * mt;
	double  d3t = 3.0 * t;

	for( int i=0; i<BxVec3::kNumVec3; i++ ) {
		diff[ i ] = 3.0 * ( -mt2                * bezierLine[ 0 ][ i ]
			          +  mt * ( 1.0 - d3t ) * bezierLine[ 1 ][ i ]
			          +  t  * ( 2.0 - d3t ) * bezierLine[ 2 ][ i ]
			          +  t2                 * bezierLine[ 3 ][ i ] );
	}

	return diff;
}


Bezier6Line::Bezier6Line( const Bezier6Line& a )
{
	for( int i=0; i<7; i++ )
		fDt[ i ] = a[ i ];
}

Bezier6Line::Bezier6Line( const BxVecAry7& a )
{
	for( int i=0; i<7; i++ )
		fDt[ i ] = a.p[ i ];
}

Bezier6Line::Bezier6Line( const BxVec3 a[ 7 ] )
{
	for( int i=0; i<7; i++ )
		fDt[ i ] = a[ i ];
}

BxVec3
Pos( const Bezier6Line& bezierLine, const double t )
{
	BxVec3  pos;

	double  t2  = t  * t;
	double  t3  = t2 * t;
	double  t4  = t3 * t;
	double  t5  = t4 * t;
	double  t6  = t5 * t;
	double  mt  = 1.0 - t;
	double  mt2 = mt  * mt;
	double  mt3 = mt2 * mt;
	double  mt4 = mt3 * mt;
	double  mt5 = mt4 * mt;
	double  mt6 = mt5 * mt;

	for( int i=0; i<BxVec3::kNumVec3; i++ ) {
		pos[ i ] =           mt6 * bezierLine[ 0 ][ i ]
			 +  6 * t  * mt5 * bezierLine[ 1 ][ i ]
			 + 15 * t2 * mt4 * bezierLine[ 2 ][ i ]
			 + 20 * t3 * mt3 * bezierLine[ 3 ][ i ]
			 + 15 * t4 * mt2 * bezierLine[ 4 ][ i ]
			 +  6 * t5 * mt  * bezierLine[ 5 ][ i ]
			 +      t6       * bezierLine[ 6 ][ i ];
	}

	return pos;
}

BxVec3
Diff( const Bezier6Line& bezierLine, const double t )
{
	BxVec3  diff;

	double  t2  = t  * t;
	double  t3  = t2 * t;
	double  t4  = t3 * t;
	double  t5  = t4 * t;
	double  mt  = 1.0 - t;
	double  mt2 = mt  * mt;
	double  mt3 = mt2 * mt;
	double  mt4 = mt3 * mt;
	double  mt5 = mt4 * mt;
	double  d2t = 2.0 * t;
	double  d3t = 3.0 * t;
	double  d6t = 6.0 * t;

	for( int i=0; i<BxVec3::kNumVec3; i++ ) {
		diff[ i ] = 6.0 * (            -mt5                 * bezierLine[ 0 ][ i ]
			          +             mt4 * ( 1.0 - d6t ) * bezierLine[ 1 ][ i ]
			          +  5.0 * t  * mt3 * ( 1.0 - d3t ) * bezierLine[ 2 ][ i ]
			          + 10.0 * t2 * mt2 * ( 1.0 - d2t ) * bezierLine[ 3 ][ i ]
			          +  5.0 * t3 * mt  * ( 2.0 - d3t ) * bezierLine[ 4 ][ i ]
			          +        t4       * ( 5.0 - d6t ) * bezierLine[ 5 ][ i ]
			          +        t5                       * bezierLine[ 6 ][ i ] );
	}

	return diff;
}



const double  NameBeziex::
GBxPi()
{
	return 3.1415926535897932;
}

const BxVec2  NameBeziex::
GBxZero2()
{
	static const BxVec2  a( 0.0, 0.0 );

	return a;
}

const BxVec3  NameBeziex::
GBxZero3()
{
	static const BxVec3  a( 0.0, 0.0, 0.0 );

	return a;
}

const double  NameBeziex::
GBxNearZero()
{
	return 1.0E-6;
}

const BxVec2  NameBeziex::
GBxNearZero2()
{
	static const BxVec2  a( 1.0E-6, 1.0E-6 );

	return a;
}

const BxVec3  NameBeziex::
GBxNearZero3()
{
	static const BxVec3  a( 1.0E-6, 1.0E-6, 1.0E-6 );

	return a;
}

const BxVec2  NameBeziex::
GBxMinusNearZero2()
{
	static const BxVec2  a( -1.0E-6, -1.0E-6 );

	return a;
}

const BxVec3  NameBeziex::
GBxMinusNearZero3()
{
	static const BxVec3  a( -1.0E-6, -1.0E-6, -1.0E-6 );

	return a;
}

const BxDir  NameBeziex::
GBxZeroDir()
{
	static const BxDir  a( 1.0, 0.0 );

	return a;
}

const BxAxisXYZ  NameBeziex::
GBxWorldAxis()
{
	static const BxAxisXYZ  a(
		BxVec3( 1.0, 0.0, 0.0 ),
		BxVec3( 0.0, 1.0, 0.0 ),
		BxVec3( 0.0, 0.0, 1.0 ) );

	return a;
}

const double  NameBeziex::
GBxNegateAngle( const double angle )
{
	return GBxSignedRegularAngle( angle + GBxPi() );
}

const double  NameBeziex::
GBxSignedRegularAngle( const double angle )
{
	double  ang2 = GBxUnsignedRegularAngle( angle );

	if( ang2 > GBxPi() )
		ang2 = ( GBxPi() * 2.0 ) - ang2;

	return ang2;
}

const double  NameBeziex::
GBxUnsignedRegularAngle( const double angle )
{
	double  ang2;

	if( angle >= 0 )
		ang2 = fmod( angle, GBxPi() * 2.0 );
	else
		ang2 = ( GBxPi() * 2.0 ) - fmod( -angle, GBxPi() * 2.0 );

	return ang2;
}


////////
